# -*- coding: utf-8 -*-
from typing import List, Dict, Tuple, Any, NamedTuple

import dataiku
from dataiku.customrecipe import get_recipe_config, get_input_names_for_role, get_output_names_for_role
from dkulib.dku_io_utils import set_column_descriptions
from dkulib.parallelizer import DataFrameParallelizer
from gpt_api_client import API_CATCH_EXCEPTIONS, GPTClient
from plugin_io_utils import ErrorHandlingEnum, validate_column_input, generate_unique
from question_answering import GPTMultiQuestionAnswerer

# ==============================================================================
# SETUP
# ==============================================================================


recipe_config = get_recipe_config()

api_configuration_preset = recipe_config.get("api_configuration_preset")
if api_configuration_preset is None or api_configuration_preset == {}:
    raise ValueError("Please specify an API configuration preset")

# text_column, input_desc and questions will trigger the following error if empty:
# <class 'ValueError'>: You must specify a valid column name
text_column = recipe_config.get("text_column")
unvalidated_questions = recipe_config.get("questions")
unvalidated_questions: List[Tuple[str, str]] = [(column_name, question) for column_name, question in
                                                unvalidated_questions.items()]

column_prefix = "gpt"
parallel_workers = api_configuration_preset.get("parallel_workers")
error_handling = (
    ErrorHandlingEnum.FAIL if get_recipe_config().get("fail_on_error") else ErrorHandlingEnum.LOG
)

# Create client
client = GPTClient(
    api_configuration_preset=api_configuration_preset,
    temperature=recipe_config.get("temperature", 0),
    max_tokens=recipe_config.get("max_tokens", 64)
)

input_dataset_name = get_input_names_for_role("input_dataset")[0]
# Not sure this check is needed
if not input_dataset_name:
    raise ValueError(
        "Cannot find input dataset. Please select an input dataset."
    )
input_dataset = dataiku.Dataset(input_dataset_name)
validate_column_input(text_column, [col["name"] for col in input_dataset.read_schema()])
input_df = input_dataset.get_dataframe()

# Make sure there is no column with no question
for (column_name, question) in unvalidated_questions:
    if not column_name:
        raise ValueError(
            "Empty column name not allowed."
        )
    if not question:
        raise ValueError(
            f"Empty question for the column {column_name}"
        )
# Generate suitable column names
questions = []
for (unvalidated_column_name, question) in unvalidated_questions:
    column_name = generate_unique(unvalidated_column_name.lower().replace(" ", "_"), input_df.columns)
    questions.append((column_name, question))
    input_df[column_name] = ""

output_dataset_name = get_output_names_for_role("output_dataset")[0]
# Not sure this check is needed
if not output_dataset_name:
    raise ValueError(
        "Cannot find output dataset. Please select an output dataset."
    )
output_dataset = dataiku.Dataset(get_output_names_for_role("output_dataset")[0])

answerer = GPTMultiQuestionAnswerer(client, questions)


def fill_values_from_response(batch: List[Dict], response: List[Any], output_column_names: NamedTuple) -> List[Dict]:
    return [
        {
            output_column_names.response: response.response_object,
            output_column_names.error_message: "",
            output_column_names.error_type: "",
            output_column_names.error_raw: "",
            **row,
            **response.parsed_answers
        }
        for response, row in zip(response, batch)
    ]


# ==============================================================================
# RUN
# ==============================================================================

df_parallelizer = DataFrameParallelizer(
    function=lambda row: answerer.get_response(input_text=str(row[text_column])),
    error_handling=error_handling,
    batch_response_parser=fill_values_from_response,
    exceptions_to_catch=API_CATCH_EXCEPTIONS,
    parallel_workers=parallel_workers,
    output_column_prefix=column_prefix,
)

df = df_parallelizer.run(input_df)

output_dataset.write_with_schema(df)

try:
    # Ugly. I feel better with this in a try catch
     gpt_desc_column_desc = {
         df_parallelizer._output_column_names._asdict()[name]: df_parallelizer.OUTPUT_COLUMN_NAME_DESCRIPTIONS[name]
         for name in df_parallelizer.OUTPUT_COLUMN_NAME_DESCRIPTIONS
     }
except:
    gpt_desc_column_desc = {}

set_column_descriptions(
    input_dataset=input_dataset,
    output_dataset=output_dataset,
    column_descriptions={
        **{k: v for k, v in questions},
        **gpt_desc_column_desc
    }
)
